#!/usr/bin/env python3

#  #####################################################################
#     Ephemetoot - A command line tool to delete your old toots
#     Copyright (C) 2018 Hugh Rundle, 2019-2020 Hugh Rundle & others
#     Initial work based on tweet-deleting script by @flesueur
#     (https://gist.github.com/flesueur/bcb2d9185b64c5191915d860ad19f23f)
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.

#     You can contact Hugh on Mastodon: @hugh@ausglam.space
#     or email: ephemetoot@hugh.run
#  #####################################################################

# import
import yaml

# from standard library
from argparse import ArgumentParser
from datetime import datetime, timezone
import os
import pkg_resources

# import funtions
from ephemetoot import ephemetoot as func

# version number from package info
vnum = pkg_resources.require("ephemetoot")[0].version

parser = ArgumentParser()
parser.add_argument(
    "--archive-deleted",
    action="store_true",
    help="Only archive toots that are being deleted",
)
parser.add_argument(
    "--config",
    action="store",
    metavar="filepath",
    default="config.yaml",
    help="Filepath of your config file, absolute or relative to the current directory. If no --config path is provided, ephemetoot will use 'config.yaml'in the current directory",
)
parser.add_argument(
    "--datestamp",
    action="store_true",
    help="Include a datetime stamp for every action (e.g. deleting a toot)",
)
parser.add_argument(
    "--hide-skipped",
    "--hide_skipped",
    action="store_true",
    help="Do not write to log when skipping saved toots",
)
parser.add_argument(
    "--init",
    action="store_true",
    help="Create a config file that is saved in the current directory",
)
parser.add_argument(
    "--pace",
    action="store_true",
    help="Slow deletion actions to match API rate limit to avoid pausing",
)
parser.add_argument(
    "-q",
    "--quiet",
    action="count",
    help="Limits logging to one line per account. Use -qq to limit logging to accounts with deleted toots and -qqq to completely suppress logging.",
)
parser.add_argument(
    "--retry-mins",
    action="store",
    metavar="minutes",
    nargs="?",
    const="1",
    default="1",
    type=int,
    help="Number of minutes to wait between retries, when an error is thrown",
)
parser.add_argument(
    "--schedule",
    action="store",
    metavar="filepath",
    nargs="?",
    const=".",
    help="Save and load plist file on MacOS",
)
parser.add_argument(
    "--test", action="store_true", help="Do a test run without deleting any toots"
)
parser.add_argument(
    "--time",
    action="store",
    metavar=("hour", "minute"),
    nargs="*",
    help="Hour and minute to schedule: e.g. 9 30 for 9.30am",
)
parser.add_argument(
    "--verbose",
    action="store_true",
    help="Log more information about errors and exceptions",
)
parser.add_argument(
    "--version",
    action="store_true",
    help="Display the version numbers of the installed and latest versions",
)

options = parser.parse_args()
if options.config[0] == "~":
    config_file = os.path.expanduser(options.config)
elif options.config[0] == "/":
    # make sure user isn't passing in something dodgy
    if os.path.exists(options.config):
        config_file = options.config
    else:
        config_file = ""
else:
    config_file = os.path.join(os.getcwd(), options.config)


def main():
    """
    Call ephemetoot.check_toots() on each user in the config file, with options set via flags from command line.
    """
    try:

        if options.init:
            func.init()
        elif options.version:
            func.version(vnum)
        elif options.schedule:
            func.schedule(options)
        else:
            if not options.quiet:
                print("")
                print("============= EPHEMETOOT v" + vnum + " ================")
                print(
                    "Running at "
                    + str(
                        datetime.now(timezone.utc).strftime("%a %d %b %Y %H:%M:%S %z")
                    )
                )
                print("================================================")
                print("")
            if options.test:
                print("This is a test run...\n")
            with open(config_file) as config:
                for accounts in yaml.safe_load_all(config):
                    for user in accounts:
                        func.check_toots(user, options)

    except FileNotFoundError as err:

        if err.filename == config_file:
            print("🕵️  Missing config file")
            print("Run \033[92mephemetoot --init\033[0m to create a new one\n")

        else:
            print("\n🤷‍♂️  The archive directory in your config file does not exist")
            print("Create the directory or correct your config before trying again\n")


if __name__ == "__main__":
    main()



import datetime
from datetime import timezone
from dateutil.tz import tzutc
import json
import os
import subprocess
import sys

import pytest
import requests

from ephemetoot import ephemetoot


########################
#         MOCKS        #
########################

toot_dict = {
    "id": 104136090490756999,
    "created_at": datetime.datetime(2020, 5, 9, 2, 17, 18, 598000, tzinfo=tzutc()),
    "in_reply_to_id": None,
    "in_reply_to_account_id": None,
    "sensitive": False,
    "spoiler_text": "",
    "visibility": "public",
    "language": "en",
    "uri": "https://example.social/users/testbot/statuses/104136090490756503",
    "url": "https://example.social/@testbot/104136090490756503",
    "replies_count": 0,
    "reblogs_count": 0,
    "favourites_count": 0,
    "favourited": False,
    "reblogged": False,
    "muted": False,
    "bookmarked": False,
    "pinned": True,
    "content": "<p>hello I am testing</p>",
    "reblog": None,
    "application": None,
    "account": {
        "id": 16186,
        "username": "testbot",
        "acct": "testbot",
        "display_name": "ephemtoot Testing Bot",
        "locked": True,
        "bot": True,
        "discoverable": False,
        "group": False,
        "created_at": datetime.datetime(
            2018, 11, 16, 23, 15, 15, 718000, tzinfo=tzutc()
        ),
        "note": "<p>Liable to explode at any time, handle with care.</p>",
        "url": "https://example.social/@testbot",
        "avatar": "https://example.social/system/accounts/avatars/000/016/186/original/66d11c4191332e7a.png?1542410869",
        "avatar_static": "https://example.social/system/accounts/avatars/000/016/186/original/66d11c4191332e7a.png?1542410869",
        "header": "https://example.social/headers/original/header.png",
        "header_static": "https://example.social/headers/original/header.png",
        "followers_count": 100,
        "following_count": 10,
        "statuses_count": 99,
        "last_status_at": datetime.datetime(2020, 8, 17, 0, 0),
        "emojis": [],
        "fields": [
            {"name": "Fully", "value": "Automated", "verified_at": None},
            {"name": "Luxury", "value": "Communism", "verified_at": None},
        ],
    },
    "media_attachments": [
      {
        "id": 123456789987654321,
        "type": "image",
        "url": "https://hugh.run/success/accomplished.jpg"
      }
    ],
    "mentions": [],
    "tags": [],
    "emojis": [],
    "card": None,
    "poll": None,
}

# Turn dict into object needed by mastodon.py
# Use this in tests after making any changes
# you need to your dict object
# NOTE: ensure values in the dict object are what you need:
# it can be mutated by any test before your test runs

def dict2obj(d):
    # checking whether object d is a
    # instance of class list
    if isinstance(d, list):
        d = [dict2obj(x) for x in d]

    # if d is not a instance of dict then
    # directly object is returned
    if not isinstance(d, dict):
        return d

    # declaring a class
    class C:
        pass

    # constructor of the class passed to obj
    obj = C()

    for k in d:
        obj.__dict__[k] = dict2obj(d[k])

    return obj


# here is our toot object - use this in tests
toot = dict2obj(toot_dict)

# config file after being parsed by yaml.safe_load
config_file = {
    "access_token": "abcd_1234",
    "username": "alice",
    "base_url": "test.social",
    "hashtags_to_keep": ["ephemetoot"],
    "days_to_keep": 14,
    "keep_pinned": True,
    "toots_to_keep": [103996285277439262, 103976473612749097, 103877521458738491],
    "visibility_to_keep": [],
    "archive": "archive",
    "archive_media": False
}

# mock GitHub API call for the version number
class MockGitHub:
    @staticmethod
    def json():
        return {"tag_name": "vLATEST_VERSION"}


# mock Mastodon
class Mocktodon:
    def __init__(self):
        return None

    def status_delete(self, t=toot):
        return None

    def status_unreblog(self, t=toot):
        return None

    def ratelimit_remaining(self):
        return 100

    def account_statuses(self, user_id=None, limit=None, max_id=None):
        # create 10 statuses
        # the first 2 will be checked in the first batch (in production it would be 40)
        user_toots = []

        def make_toot(i=1):
            if i < 11:
                keys = ("id", "created_at", "reblog", "tags", "visibility")
                vals = (
                    i,
                    datetime.datetime(2018, 11, i, 23, 15, 15, 718000, tzinfo=tzutc()),
                    False,
                    [],
                    "public",
                )
                user_toot = dict(zip(keys, vals))
                user_toots.append(user_toot)
                total = i + 1
                make_toot(total)

        user_toots.sort(reverse=True)
        make_toot(1)  # make the batch of toots
        # ignore user_id
        # filter for toots with id smaller than max_id
        this_batch = []
        # use dict_to_obj to make a toot for each toot in the obj then a list from that
        this_batch = [dict2obj(t) for t in user_toots if t["id"] > max_id][:limit]
        return this_batch


# mock argparse objects (options)
class Namespace:
    def __init__(
        self,
        archive_deleted=False,
        datestamp=False,
        hide_skipped=False,
        retry_mins=1,
        schedule=False,
        test=False,
        time=False,
        quiet=False,
    ):
        self.archive_deleted = archive_deleted
        self.datestamp = datestamp
        self.schedule = schedule
        self.time = time
        self.test = test
        self.hide_skipped = hide_skipped
        self.quiet = quiet
        self.retry_mins = retry_mins


@pytest.fixture
def mock_github_response(monkeypatch):
    def mock_get(*args, **kwargs):
        return MockGitHub()

    monkeypatch.setattr(requests, "get", mock_get)


########################
#         TESTS        #
########################

# Tests should be listed in alphabetical order
# Remember that a previous test may have mutated
# one of the values above: set all values you are using


def test_archive_toot(tmpdir):
    p = tmpdir.mkdir("archive")
    config_file["archive"] = str(p)  # make archive directory a temp test dir

    ephemetoot.archive_toot(config_file, toot)

    file_exists = os.path.exists(p + "/104136090490756999.json")
    assert file_exists

def test_archive_toot_media(tmpdir):
    p = tmpdir.mkdir("archive")
    config_file["archive"] = str(p)  # make archive directory a temp test dir
    config_file["archive_media"] = True
    ephemetoot.archive_toot_media(p, toot.media_attachments[0].url)
    image_exists = os.path.exists(p + "/hugh.run/success/accomplished.jpg")
    config_file["archive_media"] = False
    assert image_exists

def test_check_batch(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    user_id = "test_user_id"
    # limit to 2 so check_batch calls itself for the last 8 toots
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=0)
    # monkeypatch process_toot to add 1 to deleted_count and return
    # this simulates what would happen if the toot was being deleted
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: deleted_count + 1,
    )
    # run check_batch
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 10
    output = capfd.readouterr().out.split("\n")
    assert output[0] == "Removed 10 toots for alice@test.social."

def test_check_batch_quiet(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False, quiet=1)
    mastodon = Mocktodon()
    user_id = "test_user_id"
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=0)
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: deleted_count + 1,
    )
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 10
    output = capfd.readouterr().out.split("\n")
    assert output[0] == "Removed 10 toots for alice@test.social."

def test_check_batch_quiet_no_toots(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False, quiet=2)
    mastodon = Mocktodon()
    user_id = "test_user_id"
    # max_id is the last toot in our batch so this returns no toots
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=10)
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: deleted_count + 1,
    )
    # run check_batch
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 0 but with quiet=2 there should be not output
    output = capfd.readouterr().out
    assert output == ""

def test_check_batch_qq(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False, quiet=2)
    mastodon = Mocktodon()
    user_id = "test_user_id"
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=0)
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: deleted_count + 1,
    )
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 10 and message printed since there was a delete
    output = capfd.readouterr().out.split("\n")
    assert output[0] == "Removed 10 toots for alice@test.social."

def test_check_batch_qq_no_deletes(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False, quiet=2)
    mastodon = Mocktodon()
    user_id = "quiet_user_id"
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=0)
    # simulate no deletes occuring
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: 0,
    )
    # run check_batch
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 0 with no message since quiet=2
    output = capfd.readouterr().out
    assert output == ""

def test_check_batch_qqq(capfd, monkeypatch):
    config = config_file
    options = Namespace(archive_deleted=False, quiet=3)
    mastodon = Mocktodon()
    user_id = "test_user_id"
    timeline = mastodon.account_statuses(user_id=user_id, limit=2, max_id=0)
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.process_toot",
        lambda config, options, mastodon, toot, deleted_count: deleted_count + 1,
    )
    # run check_batch
    ephemetoot.check_batch(config, options, mastodon, user_id, timeline, 0)
    # deleted_count should be 10 and no message should be printed since quiet=3
    output = capfd.readouterr().out
    assert output == ""

def test_console_print(capfd):
    ephemetoot.console_print(
        "test123", Namespace(test=False, hide_skipped=False, quiet=False), False
    )
    assert capfd.readouterr().out == "test123\n"


def test_console_print_quiet():
    result = ephemetoot.console_print(
        "test123", Namespace(test=False, hide_skipped=False, quiet=True), False
    )
    assert result == None


def test_console_print_skip():
    result = ephemetoot.console_print(
        "test123", Namespace(test=False, hide_skipped=True, quiet=False), True
    )
    assert result == None


def test_datestamp_now():
    datestamp = ephemetoot.datestamp_now()
    date_object = datetime.datetime.strptime(datestamp, "%a %d %b %Y %H:%M:%S %z")
    # use timetuple() to exclude differences in milliseconds
    assert datetime.datetime.now(timezone.utc).timetuple() == date_object.timetuple()


def test_init(monkeypatch, tmpdir):

    # monkeypatch current directory
    current_dir = tmpdir.mkdir("current_dir")  # temporary directory for testing
    monkeypatch.chdir(current_dir)

    # monkeypatch input ...outputs
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.compulsory_input", lambda a, b, c: "compulsory"
    )
    monkeypatch.setattr("ephemetoot.ephemetoot.digit_input", lambda a, b, c: "14")
    monkeypatch.setattr("ephemetoot.ephemetoot.yes_no_input", lambda a, b: "false")
    monkeypatch.setattr(
        "ephemetoot.ephemetoot.optional_input", lambda a, b, c: "optional"
    )

    # run init
    ephemetoot.init()
    assert os.path.exists(os.path.join(current_dir, "config.yaml"))


def test_init_archive_path(tmpdir):

    good_path = tmpdir.mkdir("archive_dir")  # temporary directory for testing
    wrong = ephemetoot.sanitise_input(
        os.path.join(good_path, "/bad/path/"), "Archive path", None
    )
    ok = ephemetoot.sanitise_input(good_path, "Archive path", None)

    assert ok == "ok"
    assert wrong == "That directory does not exist, please try again"


def test_init_sanitise_id_list():
    tags = ("\033[96m", "\033[2m", "\033[0m")
    wrong = ephemetoot.sanitise_input(
        "987654321, toot_id_number", "Toots to keep", tags
    )
    also_wrong = ephemetoot.sanitise_input("toot_id_number", "Toots to keep", tags)
    ok = ephemetoot.sanitise_input("1234598745, 999933335555", "Toots to keep", tags)
    also_ok = ephemetoot.sanitise_input("1234598745", "Toots to keep", tags)

    assert wrong == "Toot IDs must be numeric and separated with commas"
    assert also_wrong == "Toot IDs must be numeric and separated with commas"
    assert ok == "ok"
    assert also_ok == "ok"


def test_init_sanitise_tag_list():
    tags = ("\033[96m", "\033[2m", "\033[0m")
    wrong = ephemetoot.sanitise_input("#tag, another_tag", "Hashtags to keep", tags)
    also_wrong = ephemetoot.sanitise_input("tag, another tag", "Hashtags to keep", tags)
    still_wrong = ephemetoot.sanitise_input("tag, 12345", "Hashtags to keep", tags)
    ok = ephemetoot.sanitise_input("tag123, another_TAG", "Hashtags to keep", tags)
    also_ok = ephemetoot.sanitise_input("single_tag", "Hashtags to keep", tags)

    error = (
        "Hashtags must not include '#' and must match rules at "
        + tags[0]
        + "https://docs.joinmastodon.org/user/posting/#hashtags"
        + tags[2]
    )

    assert ok == "ok"
    assert also_ok == "ok"
    assert wrong == error
    assert also_wrong == error
    assert still_wrong == error


def test_init_sanitise_url():
    tags = ("\033[96m", "\033[2m", "\033[0m")
    wrong = ephemetoot.sanitise_input("http://example.social", "Base URL", tags)
    ok = ephemetoot.sanitise_input("example.social", "Base URL", tags)

    assert (
        wrong
        == "Provide full domain without protocol prefix (e.g. \033[2mexample.social\033[0m, not \033[2mhttp://example.social\033[0m)"
    )
    assert ok == "ok"


def test_init_sanitise_username():
    tags = ("\033[96m", "\033[2m", "\033[0m")
    wrong = ephemetoot.sanitise_input("@alice", "Username", tags)
    ok = ephemetoot.sanitise_input("alice", "Username", tags)

    assert wrong == "Do not include '@' in username, please try again"
    assert ok == "ok"


def test_init_sanitise_visibility_list():
    tags = ("\033[96m", "\033[2m", "\033[0m")
    wrong = ephemetoot.sanitise_input("nonexistent", "Visibility to keep", tags)
    also_wrong = ephemetoot.sanitise_input("direct public", "Visibility to keep", tags)
    ok = ephemetoot.sanitise_input("direct", "Visibility to keep", tags)
    also_ok = ephemetoot.sanitise_input("direct, public", "Visibility to keep", tags)

    error = (
        "Valid values are one or more of 'public', 'unlisted', 'private' or 'direct'"
    )
    assert ok == "ok"
    assert also_ok == "ok"
    assert wrong == error
    assert also_wrong == error


def test_jsondefault():
    d = ephemetoot.jsondefault(toot.created_at)
    assert d == "2020-05-09T02:17:18.598000+00:00"


def test_process_toot(capfd, tmpdir, monkeypatch):
    # config uses config_listed at top of this tests file
    p = tmpdir.mkdir("archive")  # use temporary test directory
    config_file["archive"] = str(p)
    config_file["keep_pinned"] = False
    config_file["toots_to_keep"] = []
    config_file["visibility_to_keep"] = []
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    toot_dict["pinned"] = False
    toot_dict["visibility"] = "public"
    toot_dict["reblog"] = False
    toot = dict2obj(toot_dict)
    ephemetoot.process_toot(config_file, options, mastodon, toot, 0)
    assert (
        capfd.readouterr().out
        == "❌ deleting toot 104136090490756999 tooted 09 May 2020\n"
    )


def test_process_toot_pinned(capfd, tmpdir):
    # config uses config_listed at top of this tests file
    p = tmpdir.mkdir("archive")  # use temporary test directory
    config_file["archive"] = str(p)
    config_file["keep_pinned"] = True
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    toot_dict["pinned"] = True
    toot = dict2obj(toot_dict)
    ephemetoot.process_toot(config_file, options, mastodon, toot, 0)
    assert capfd.readouterr().out == "📌 skipping pinned toot - 104136090490756999\n"


def test_process_toot_saved(capfd, tmpdir):
    # config uses config_listed at top of this tests file
    p = tmpdir.mkdir("archive")  # use temporary test directory
    config_file["archive"] = str(p)
    config_file["keep_pinned"] = False
    config_file["toots_to_keep"].append(104136090490756999)
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    toot_dict["pinned"] = False
    toot = dict2obj(toot_dict)
    ephemetoot.process_toot(config_file, options, mastodon, toot, 0)
    assert capfd.readouterr().out == "💾 skipping saved toot - 104136090490756999\n"


def test_process_toot_visibility(capfd, tmpdir):
    # config uses config_listed at top of this tests file
    p = tmpdir.mkdir("archive")  # use temporary test directory
    config_file["archive"] = str(p)
    config_file["keep_pinned"] = False  # is true above so make false
    config_file["toots_to_keep"].remove(104136090490756999)  # don't keep this toot
    config_file["visibility_to_keep"].append("testing")
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    toot_dict["pinned"] = False  # is true above so make false
    toot_dict["visibility"] = "testing"
    toot = dict2obj(toot_dict)
    ephemetoot.process_toot(config_file, options, mastodon, toot, 0)
    assert capfd.readouterr().out == "👀 skipping testing toot - 104136090490756999\n"


def test_process_toot_hashtag(capfd, tmpdir, monkeypatch):
    # config uses config_listed at top of this tests file
    p = tmpdir.mkdir("archive")  # use temporary test directory
    config_file["archive"] = str(p)
    config_file["keep_pinned"] = False
    config_file["toots_to_keep"] = []
    config_file["visibility_to_keep"] = []
    options = Namespace(archive_deleted=False)
    mastodon = Mocktodon()
    toot_dict["pinned"] = False
    toot_dict["visibility"] = "public"
    toot_dict["reblog"] = True
    toot = dict2obj(toot_dict)

    ephemetoot.process_toot(config_file, options, mastodon, toot, 0)
    assert (
        capfd.readouterr().out
        == "👎 unboosting toot 104136090490756999 boosted 09 May 2020\n"
    )


def test_retry_on_error():
    # Namespace object constructed from top of tests (representing options)
    # toot comes from variable at top of test
    mastodon = Mocktodon()
    toot = dict2obj(toot_dict)
    retry = ephemetoot.retry_on_error(Namespace(retry_mins=True), mastodon, toot, 5)
    assert retry == None  # should not return an error


def test_retry_on_error_max_tries():
    # Namespace object constructed from top of tests (representing options)
    # toot and mastodon come from objects at top of test
    with pytest.raises(TimeoutError):
        mastodon = Mocktodon()
        toot = dict2obj(toot_dict)
        retry = ephemetoot.retry_on_error(Namespace(retry_mins=True), mastodon, toot, 7)


def test_schedule(monkeypatch, tmpdir):

    home = tmpdir.mkdir("current_dir")  # temporary directory for testing
    launch = tmpdir.mkdir("TestAgents")  # temporary directory for testing

    # monkeypatch directories and suppress the plist loading process
    # NOTE: it may be possible to test the plist loading process
    # but I can't work out how to do it universally / consistently

    def mock_current_dir():
        return str(home)

    def mock_home_dir_expansion(arg):
        return str(launch)

    def suppress_subprocess(args, stdout=None, stderr=None, shell=None):
        return None

    monkeypatch.setattr(os, "getcwd", mock_current_dir)
    monkeypatch.setattr(os.path, "expanduser", mock_home_dir_expansion)
    monkeypatch.setattr(subprocess, "run", suppress_subprocess)

    # now we run the function we're testing
    ephemetoot.schedule(Namespace(schedule=".", time=None))

    # assert the plist file was created
    plist_file = os.path.join(launch, "ephemetoot.scheduler.plist")
    assert os.path.lexists(plist_file)

    # check that correct values were modified in the file
    f = open(plist_file, "r")
    plist = f.readlines()
    assert plist[7] == "  <string>" + str(home) + "</string>\n"
    assert plist[7] == "  <string>" + str(home) + "</string>\n"
    assert plist[10] == "    <string>" + sys.argv[0] + "</string>\n"
    assert plist[12] == "    <string>" + str(home) + "/config.yaml</string>\n"
    assert plist[15] == "  <string>" + str(home) + "/ephemetoot.log</string>\n"
    assert plist[17] == "  <string>" + str(home) + "/ephemetoot.error.log</string>\n"


def test_schedule_with_time(monkeypatch, tmpdir):

    home = tmpdir.mkdir("current_dir")  # temporary directory for testing
    launch = tmpdir.mkdir("TestAgents")  # temporary directory for testing

    # monkeypatch directories and suppress the plist loading process
    # NOTE: it may be possible to test the plist loading process
    # but I can't work out how to do it universally / consistently

    def mock_current_dir():
        return str(home)

    def mock_home_dir_expansion(arg):
        return str(launch)

    def suppress_subprocess(args, stdout=None, stderr=None, shell=None):
        return None

    monkeypatch.setattr(os, "getcwd", mock_current_dir)
    monkeypatch.setattr(os.path, "expanduser", mock_home_dir_expansion)
    monkeypatch.setattr(subprocess, "run", suppress_subprocess)

    # now we run the function we're testing
    ephemetoot.schedule(Namespace(schedule=".", time=["10", "30"]))

    # assert the plist file was created
    plist_file = os.path.join(launch, "ephemetoot.scheduler.plist")
    assert os.path.lexists(plist_file)

    # assert that correct values were modified in the file
    f = open(plist_file, "r")
    plist = f.readlines()

    assert plist[21] == "    <integer>10</integer>\n"
    assert plist[23] == "    <integer>30</integer>\n"


def test_tooted_date():
    string = ephemetoot.tooted_date(toot)
    created = datetime.datetime(2020, 5, 9, 2, 17, 18, 598000, tzinfo=timezone.utc)
    test_string = created.strftime("%d %b %Y")

    assert string == test_string


def test_version(mock_github_response, capfd):
    ephemetoot.version("TEST_VERSION")
    output = capfd.readouterr().out
    msg = """
ephemetoot ==> 🥳 ==> 🧼 ==> 😇
-------------------------------
You are using release: \033[92mvTEST_VERSION\033[0m
The latest release is: \033[92mvLATEST_VERSION\033[0m
To upgrade to the most recent version run \033[92mpip install --upgrade ephemetoot\033[0m\n"""

    assert output == msg



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ephemetoot']

package_data = \
{'': ['*']}

install_requires = \
['mastodon.py>=1.4.3,<2.0.0', 'pyyaml>=5.0,<6.0', 'requests>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['ephemetoot = ephemetoot.console:main']}

setup_kwargs = {
    'name': 'ephemetoot',
    'version': '3.1.2',
    'description': 'A command line tool to delete your old toots',
    'long_description': '# 🥳 ==> 🧼 ==> 😇\n\n## Prior work\nThe initial `ephemetoot` script was based on [this tweet-deleting script](https://gist.github.com/flesueur/bcb2d9185b64c5191915d860ad19f23f) by [@flesueur](https://github.com/flesueur)\n\n`ephemetoot` relies heavily on the [Mastodon.py](https://pypi.org/project/Mastodon.py/) package by [@halcy](https://github.com/halcy)\n\n## Usage\n\nYou can use `ephemetoot` to delete [Mastodon](https://github.com/tootsuite/mastodon) toots that are older than a certain number of days (default is 365). Toots can optionally be saved from deletion if:\n* they are pinned; or\n* they include certain hashtags; or\n* they have certain visibility; or\n* they are individually listed to be kept\n\nThere are various options controlling timing, scheduling, and output.\n\nRun from the command line with `ephemetoot`.\n\nRun `ephemetoot --help` or read the docs for all options.\n\n## Contributing\n\nephemetoot is packaged using `poetry` and tested using `pytest`.\n\nFor all bugs, suggestions, pull requests or other contributions, please check the [contributing guide](https://github.com/hughrun/ephemetoot/blob/master/docs/contributing.md).\n',
    'author': 'Hugh Rundle',
    'author_email': 'ephemetoot@hugh.run',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://ephemetoot.hugh.run',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# standard library
from datetime import date, datetime, timedelta, timezone
import json
import os
import re
import urllib.parse
import subprocess
import sys
import time

# third party
from mastodon import (
    Mastodon,
    MastodonError,
    MastodonAPIError,
    MastodonNetworkError,
    MastodonRatelimitError,
)
import requests

# local
from ephemetoot import plist


def compulsory_input(tags, name, example):

    value = ""
    while len(value) < 1:
        if example:
            value = input(tags[0] + name + tags[1] + example + tags[2])
        else:
            value = input(tags[0] + name + tags[2])

        sanitised = sanitise_input(value, name, tags)

        if len(value) > 0 and (sanitised == "ok" or sanitised == None):
            return value
        else:
            if len(value) > 0 and sanitised != None:
                print(sanitised)
            value = ""


def digit_input(tags, name, example):

    value = ""
    while value.isdigit() == False:
        if example:
            value = input(tags[0] + name + tags[1] + example + tags[2])
        else:
            value = input(tags[0] + name + tags[2])

    return value


def yes_no_input(tags, name):
    value = ""
    while value not in ["y", "n"]:
        value = input(tags[0] + name + tags[1] + "(y or n):" + tags[2])
    return_val = "true" if value == "y" else "false"
    return return_val


def optional_input(tags, name, example):

    incomplete = True
    while incomplete:
        value = input(tags[0] + name + tags[1] + example + tags[2])
        sanitised = sanitise_input(value, name, tags)
        if len(value) > 0 and (sanitised == "ok" or sanitised == None):
            incomplete = False
            return value
        elif len(value) > 0 and sanitised != None:
            print(sanitised)
        else:
            return ""


def sanitise_input(value, input_type, tags):
    """
    Check that data entered when running --init complies with requirements
    """

    if input_type == "Username":
        return (
            "Do not include '@' in username, please try again"
            if value.startswith("@")
            else "ok"
        )

    if input_type == "Base URL":
        error = value.startswith("http") or value.find(".") == -1
        return (
            "Provide full domain without protocol prefix (e.g. "
            + tags[1]
            + "example.social"
            + tags[2]
            + ", not "
            + tags[1]
            + "http://example.social"
            + tags[2]
            + ")"
            if error
            else "ok"
        )

    if input_type == "Toots to keep":
        l = value.split(",")

        def check(s):
            d = s.strip()
            if not d.isdigit():
                return False

        allnum = map(check, l)
        return (
            "Toot IDs must be numeric and separated with commas"
            if False in list(allnum)
            else "ok"
        )

    if input_type == "Hashtags to keep":
        l = value.split(",")

        def check(s):
            d = s.strip()
            if d.isdigit():
                return False
            if not re.fullmatch(r"[\w]+", d, flags=re.IGNORECASE):
                return False

        complies = map(check, l)
        return_string = (
            "Hashtags must not include '#' and must match rules at "
            + tags[0]
            + "https://docs.joinmastodon.org/user/posting/#hashtags"
            + tags[2]
        )
        return return_string if False in list(complies) else "ok"

    if input_type == "Visibility to keep":
        l = value.split(",")
        viz_options = set(["public", "unlisted", "private", "direct"])

        def check(s):
            d = [s.strip().lower()]
            intersects = viz_options.intersection(d)
            if len(intersects) == 0:
                return False

        complies = map(check, l)
        return_string = "Valid values are one or more of 'public', 'unlisted', 'private' or 'direct'"
        return return_string if False in list(complies) else "ok"

    if input_type == "Archive path":
        path = (
            os.path.expanduser(value)
            if len(str(value)) > 0 and str(value)[0] == "~"
            else value
        )
        response = (
            "ok"
            if os.path.exists(path)
            else "That directory does not exist, please try again"
        )
        return response


def init():
    """
    Creates a config.yaml file in the current directory, based on user input.
    """
    try:

        # text colour markers (beginning, example, end)
        tags = ("\033[96m", "\033[2m", "\033[0m")

        print("\nCreate your config.yaml file.\n")
        print(
            "For help check out the docs at ",
            tags[0],
            "ephemetoot.hugh.run",
            tags[2],
            "\n",
            sep="",
        )

        conf_token = compulsory_input(tags, "Access token: ", None)
        conf_user = compulsory_input(
            tags, "Username", "(without the '@' - e.g. alice):"
        )
        conf_url = compulsory_input(tags, "Base URL", "(e.g. example.social):")
        conf_days = digit_input(tags, "Days to keep", "(default 365):")
        conf_pinned = yes_no_input(tags, "Keep pinned toots?")
        conf_keep_toots = optional_input(
            tags, "Toots to keep", "(optional list of IDs separated by commas):"
        )
        conf_keep_hashtags = optional_input(
            tags,
            "Hashtags to keep",
            "(optional list without '#' e.g. mastodon, gardening, cats):",
        )
        conf_keep_visibility = optional_input(
            tags, "Visibility to keep", "(optional list e.g. 'direct'):"
        )
        conf_archive = optional_input(
            tags, "Archive path", "(optional filepath for archive):"
        )

        if len(conf_archive) > 0:
            conf_archive_media = yes_no_input(tags, "Archive media?")

        # write out the config file
        with open("config.yaml", "w") as configfile:

            configfile.write("-")
            configfile.write("\n  access_token: " + conf_token)
            configfile.write("\n  username: " + conf_user)
            configfile.write("\n  base_url: " + conf_url)
            configfile.write("\n  days_to_keep: " + conf_days)
            configfile.write("\n  keep_pinned: " + conf_pinned)

            if len(conf_keep_toots) > 0:
                keep_list = conf_keep_toots.split(",")
                configfile.write("\n  toots_to_keep:")
                for toot in keep_list:
                    configfile.write("\n    - " + toot.strip())

            if len(conf_keep_hashtags) > 0:
                tag_list = conf_keep_hashtags.split(",")
                configfile.write("\n  hashtags_to_keep:")
                for tag in tag_list:
                    configfile.write("\n    - " + tag.strip())

            if len(conf_keep_visibility) > 0:
                viz_list = conf_keep_visibility.split(",")
                configfile.write("\n  visibility_to_keep:")
                for mode in viz_list:
                    configfile.write("\n    - " + mode.strip())

            if len(conf_archive) > 0:
                configfile.write("\n  archive: " + conf_archive)
                configfile.write("\n  archive_media: " + conf_archive_media)

            configfile.close()

    except Exception as e:
        print(e)


def version(vnum):
    """
    Prints current and latest version numbers to console.
    """

    try:
        latest = requests.get(
            "https://api.github.com/repos/hughrun/ephemetoot/releases/latest"
        )
        res = latest.json()
        latest_version = res["tag_name"]
        print("\nephemetoot ==> 🥳 ==> 🧼 ==> 😇")
        print("-------------------------------")
        print("You are using release: \033[92mv", vnum, "\033[0m", sep="")
        print("The latest release is: \033[92m" + latest_version + "\033[0m")
        print(
            "To upgrade to the most recent version run \033[92mpip install --upgrade ephemetoot\033[0m"
        )

    except Exception as e:
        print("Something went wrong:", e)


def schedule(options):

    """
    Creates and loads a plist file for scheduled running with launchd. If --time flag is used, the scheduled time is set accordingly. Note that this is designed for use on MacOS.
    """
    try:

        if options.schedule == ".":
            working_dir = os.getcwd()
        else:
            working_dir = options.schedule

        lines = plist.default_file.splitlines()
        lines[7] = "  <string>" + working_dir + "</string>"
        lines[10] = "    <string>" + sys.argv[0] + "</string>"
        lines[12] = "    <string>" + working_dir + "/config.yaml</string>"
        lines[15] = "  <string>" + working_dir + "/ephemetoot.log</string>"
        lines[17] = "  <string>" + working_dir + "/ephemetoot.error.log</string>"

        if options.time:
            lines[21] = "    <integer>" + options.time[0] + "</integer>"
            lines[23] = "    <integer>" + options.time[1] + "</integer>"

        # write out file directly to ~/Library/LaunchAgents
        f = open(
            os.path.join(
                os.path.expanduser("~/Library/LaunchAgents"),
                "ephemetoot.scheduler.plist",
            ),
            mode="w",
        )
        for line in lines:
            if line == lines[-1]:
                f.write(line)
            else:
                f.write(line + "\n")
        f.close()
        sys.tracebacklimit = 0  # suppress Tracebacks
        # unload any existing file (i.e. if this is an update to the file) and suppress any errors
        subprocess.run(
            ["launchctl unload ~/Library/LaunchAgents/ephemetoot.scheduler.plist"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            shell=True,
        )
        # load the new file
        subprocess.run(
            ["launchctl load ~/Library/LaunchAgents/ephemetoot.scheduler.plist"],
            shell=True,
        )
        print("⏰ Scheduled!")
    except Exception as e:
        print("🙁 Scheduling failed.", e)
        if options.verbose:
            print(e)


def archive_toot_media(archive_path, full_url):
    url = urllib.parse.urlparse(full_url)
    (dir_name, file_name) = os.path.split(url.path)
    media_archive_path = os.path.join(archive_path, url.netloc, dir_name[1:])
    media_archive_file_path = os.path.join(media_archive_path, file_name)
    if os.path.isfile(media_archive_file_path):
        return
    os.makedirs(media_archive_path, exist_ok=True)
    r = requests.get(full_url)
    with open(media_archive_file_path, "wb") as f:
        f.write(r.content)


def archive_toot(config, toot):
    archive_media = "archive_media" in config and config["archive_media"]

    # define archive path
    if config["archive"][0] == "~":
        archive_path = os.path.expanduser(config["archive"])
    elif config["archive"][0] == "/":
        archive_path = config["archive"]
    else:
        archive_path = os.path.join(os.getcwd(), config["archive"])
    if archive_path[-1] != "/":
        archive_path += "/"

    filename = os.path.join(archive_path, str(toot.id) + ".json")

    # write to file
    with open(filename, "w") as f:
        f.write(json.dumps(toot, indent=4, default=jsondefault))
        f.close()

    if archive_media and "media_attachments" in toot:
        for media_attachment in toot["media_attachments"]:
            if "url" in media_attachment:
                archive_toot_media(archive_path, media_attachment["url"])


def jsondefault(obj):
    if isinstance(obj, (date, datetime)):
        return obj.isoformat()


def tooted_date(toot):
    return toot.created_at.strftime("%d %b %Y")


def datestamp_now():
    return str(datetime.now(timezone.utc).strftime("%a %d %b %Y %H:%M:%S %z"))


def console_print(msg, options, skip):

    skip_announcement = True if (options.hide_skipped and skip) else False
    if not (skip_announcement or options.quiet):

        if options.datestamp:
            msg = datestamp_now() + " : " + msg

        print(msg)


def print_rate_limit_message(mastodon):

    now = time.time()
    diff = mastodon.ratelimit_reset - now

    print(
        "\nRate limit reached at",
        datestamp_now(),
        "- next reset due in",
        str(format(diff / 60, ".0f")),
        "minutes.\n",
    )


def retry_on_error(options, mastodon, toot, attempts=0):

    if attempts < 6:
        try:
            console_print(
                "Attempt " + str(attempts) + " at " + datestamp_now(), options, False
            )
            mastodon.status_delete(toot)
        except:
            attempts += 1
            time.sleep(60 * options.retry_mins)
            retry_on_error(options, mastodon, toot, attempts)
    else:
        raise TimeoutError("Gave up after 5 attempts")


def process_toot(config, options, mastodon, toot, deleted_count=0):

    keep_pinned = "keep_pinned" in config and config["keep_pinned"]
    toots_to_keep = config["toots_to_keep"] if "toots_to_keep" in config else []
    visibility_to_keep = (
        config["visibility_to_keep"] if "visibility_to_keep" in config else []
    )
    hashtags_to_keep = (
        set(config["hashtags_to_keep"]) if "hashtags_to_keep" in config else set()
    )
    days_to_keep = config["days_to_keep"] if "days_to_keep" in config else 365
    cutoff_date = datetime.now(timezone.utc) - timedelta(days=days_to_keep)

    if toot.id and "archive" in config:

        if not options.archive_deleted:
            # write toot to archive
            archive_toot(config, toot)

    toot_tags = set()
    for tag in toot.tags:
        toot_tags.add(tag.name)

    try:
        if keep_pinned and hasattr(toot, "pinned") and toot.pinned:
            console_print("📌 skipping pinned toot - " + str(toot.id), options, True)

        elif toot.id in toots_to_keep:
            console_print("💾 skipping saved toot - " + str(toot.id), options, True)

        elif toot.visibility in visibility_to_keep:
            console_print(
                "👀 skipping " + toot.visibility + " toot - " + str(toot.id),
                options,
                True,
            )

        elif len(hashtags_to_keep.intersection(toot_tags)) > 0:
            console_print(
                "#️⃣  skipping toot with hashtag - " + str(toot.id), options, True
            )

        elif cutoff_date > toot.created_at:
            if hasattr(toot, "reblog") and toot.reblog:
                console_print(
                    "👎 unboosting toot "
                    + str(toot.id)
                    + " boosted "
                    + tooted_date(toot),
                    options,
                    False,
                )

                deleted_count += 1
                # unreblog the original toot (their toot), not the toot created by boosting (your toot)
                if not options.test:
                    if mastodon.ratelimit_remaining == 0:
                        console_print(
                            "Rate limit reached. Waiting for a rate limit reset",
                            options,
                            False,
                        )

                    # check for --archive-deleted
                    if options.archive_deleted and "id" in toot and "archive" in config:
                        # write toot to archive
                        archive_toot(config, toot)

                    mastodon.status_unreblog(toot.reblog)

            else:
                console_print(
                    "❌ deleting toot " + str(toot.id) + " tooted " + tooted_date(toot),
                    options,
                    False,
                )

                deleted_count += 1
                time.sleep(
                    2
                )  # wait 2 secs between deletes to be a bit nicer to the server

                if not options.test:
                    # deal with rate limits
                    if mastodon.ratelimit_remaining == 0 and not options.quiet:
                        print_rate_limit_message(mastodon)

                    # check for --archive-deleted
                    if options.archive_deleted and "id" in toot and "archive" in config:
                        archive_toot(config, toot)

                    # finally we actually delete the toot
                    mastodon.status_delete(toot)

        # return the deleted_count back so that it can be tallied within check_batch()
        return deleted_count

    except MastodonRatelimitError:

        print_rate_limit_message(mastodon)
        time.sleep(diff + 1)  # wait for rate limit to reset

    # If a server goes offline for maintenance etc halfway through a run, we don't necessarily
    # want to just error out. Handling it here allows us to give it time to sort itself out.
    except MastodonError as e:

        if options.verbose:
            print("🛑 ERROR deleting toot -", str(toot.id), "\n", e)
        else:
            print(
                "🛑 ERROR deleting toot -",
                str(toot.id),
                "-",
                str(e.args[0]),
                "-",
                str(e.args[3]),
            )

        console_print(
            "Waiting " + str(options.retry_mins) + " minutes before re-trying",
            options,
            False,
        )
        time.sleep(60 * options.retry_mins)
        retry_on_error(options, mastodon, toot, attempts=2)


def check_batch(config, options, mastodon, user_id, timeline, deleted_count=0):
    """
    Check a batch of up to 40 toots. This is usually triggered by check_toots, and then recursively calls itself until all toots within the time period specified have been checked.
    """

    try:
        for toot in timeline:
            # process_toot returns the value of the deleted_count so we can keep track here
            deleted_count = process_toot(config, options, mastodon, toot, deleted_count)

        # the account_statuses call is paginated with a 40-toot limit
        # get the id of the last toot to include as 'max_id' in the next API call.
        # then keep triggering new rounds of check_toots() until there are no more toots to check
        max_id = timeline[-1:][0].id
        next_batch = mastodon.account_statuses(user_id, limit=40, max_id=max_id)
        if len(next_batch) > 0:
            check_batch(config, options, mastodon, user_id, next_batch, deleted_count)
        else:
            if not options.test:
                if options.datestamp:
                    print("\n", datestamp_now(), end=" : ")

                # options.quiet can be None
                if (
                    (not options.quiet)
                    or options.quiet <= 1
                    or (options.quiet == 2 and deleted_count)
                ):
                    print(
                        "Removed "
                        + str(deleted_count)
                        + " toots for "
                        + config["username"]
                        + "@"
                        + config["base_url"]
                        + ".\n"
                    )

                if not options.quiet:
                    print("---------------------------------------")
                    print("🥳 ==> 🧼 ==> 😇 User cleanup complete!")
                    print("---------------------------------------\n")

            else:

                if options.quiet:
                    if options.datestamp:
                        print("\n", datestamp_now(), sep="", end=" : ")

                    print(
                        "Test run completed. This would have removed",
                        str(deleted_count),
                        "toots.\n",
                    )

                else:
                    print("---------------------------------------")
                    print("🥳 ==> 🧪 ==> 📋 Test run complete!")
                    print("This would have removed", str(deleted_count), "toots.")
                    print("---------------------------------------\n")

    except IndexError:
        if not options.quiet or options.quiet <= 1:
            print(
                "No toots found for "
                + config["username"]
                + "@"
                + config["base_url"]
                + ".\n"
            )


def check_toots(config, options, retry_count=0):
    """
    The main function, uses the Mastodon API to check all toots in the user timeline, and delete any that do not meet any of the exclusion criteria from the config file.
    """
    try:
        if not options.quiet:
            print(
                "Fetching account details for @",
                config["username"],
                "@",
                config["base_url"],
                sep="",
            )

        if options.pace:
            mastodon = Mastodon(
                access_token=config["access_token"],
                api_base_url="https://" + config["base_url"],
                ratelimit_method="pace",
            )
        else:
            mastodon = Mastodon(
                access_token=config["access_token"],
                api_base_url="https://" + config["base_url"],
                ratelimit_method="wait",
            )

        user_id = mastodon.account_verify_credentials().id  # verify user and get ID
        account = mastodon.account(user_id)  # get the account
        timeline = mastodon.account_statuses(user_id, limit=40)  # initial batch

        if not options.quiet:
            print("Checking", str(account.statuses_count), "toots")

        # check first batch
        # check_batch() then recursively keeps looping until all toots have been checked
        check_batch(config, options, mastodon, user_id, timeline)

    except KeyboardInterrupt:
        print("Operation aborted.")

    except KeyError as val:
        print("\n⚠️  error with in your config.yaml file!")
        print("Please ensure there is a value for " + str(val) + "\n")

    except MastodonAPIError as e:
        if e.args[1] == 401:
            print(
                "\n🙅  User and/or access token does not exist or has been deleted (401)\n"
            )
        elif e.args[1] == 404:
            print("\n🔭  Can't find that server (404)\n")
        else:
            print("\n😕  Server has returned an error (5xx)\n")

        if options.verbose:
            print(e, "\n")

    except MastodonNetworkError as e:
        if retry_count == 0:
            print("\n📡  ephemetoot cannot connect to the server - are you online?")
            if options.verbose:
                print(e)
        if retry_count < 4:
            print("Waiting " + str(options.retry_mins) + " minutes before trying again")
            time.sleep(60 * options.retry_mins)
            retry_count += 1
            print("Attempt " + str(retry_count + 1))
            check_toots(config, options, retry_count)
        else:
            print("Gave up waiting for network\n")

    except Exception as e:
        if options.verbose:
            print("ERROR:", e)
        else:
            print("ERROR:", str(e.args[0]), "\n")
